// ----
// ---- file   : SyModuleVST.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2010 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 06Oct2010, 07Oct2010, 10Oct2010, 11Oct2010, 16Oct2010
// ----
// ----
// ----

#define YAC_EPSILONCOMPARE_ABS defined
#define YAC_EPSILONCOMPARE_ABS_DEFAULT defined

#include "tkvst2.h"

#include "../tksynth/tksynth.h"

#include "VST2Plugin.h"
#include "VST2PinProperties.h"

#include "SyModuleVST.h"


static sBool IsEmptyString(YAC_Value *rLabel) {
   if(rLabel->type >= YAC_TYPE_OBJECT)
   {
      if(YAC_Is_String(rLabel->value.object_val))
      {
         YAC_String *s = (YAC_String*) rLabel->value.object_val;
         if(NULL != s->chars)
         {
            return (s->length <= 1);
         }
      }
   }
   return 1;
}


SyModuleVST::SyModuleVST(void) {
   plugin = NULL;

   ui_vstgenerator_id = 0;

   patch_data_type = SYMODULEVST_DATATYPE_NONE;
   patch_data      = NULL;
}

SyModuleVST::~SyModuleVST() {
   unloadVST();

   YAC_DELETE_SAFE(patch_data);
}

void SyModuleVST::unloadVST(void) {

   if(NULL != plugin)
   {
      // Stop processing
      plugin->stopProcess();

      // Suspend
      plugin->suspend();

      // Close
      //  (note) close is called in plugin destructor
      ////plugin->close();

      YAC_DELETE_SAFE(plugin);
   }
}

sBool SyModuleVST::initVST(YAC_String *_pathName, sF32 _rate, sUI _maxFrames) {

   // (note) usually this is called when the module is uninitialized and I/O ports need to be allocated
   //        however, when restoring this module, the I/O ports have already been allocated in readStream()
   //        and the port object addresses must NOT change anymore

   // true during delayed loading of VST plugins in ST2
   sBool bPrealloc = (num_inputs > 0) || (num_outputs > 0);

   if(YAC_Is_String(_pathName))
   {
      printf("xxx initVST 1\n");
      allocPlugin();

      printf("xxx initVST 2\n");

      if(plugin->loadPlugin(_pathName, 0/*selectShellPluginUID*/))
      {
         // OK, plugin loaded, now determine the number of required tksynth I/O ports
         printf("xxx initVST 3\n");

         plugin->open();
         plugin->setSampleRate(_rate);
         plugin->setBlockSize(_maxFrames);

         printf("xxx initVST 4\n");

         // Query input ports
         sSI numIn  = plugin->getNumInputs();
         sSI numIPorts = 0;

         sSI inIdx = 0;

         if(bPrealloc)
         {
            numIPorts = (sSI) num_inputs;
         }
         else
         {
            // Determine number of tksynth input ports
            while(inIdx < numIn)
            {
               // Query input pin
               YAC_Value r;
            
               plugin->getInputProperties((sUI)inIdx, &r);

               if(r.type >= YAC_TYPE_OBJECT)
               {
                  VST2PinProperties *pin = (VST2PinProperties *) r.value.object_val;

                  if(pin->isStereo())
                  {
                     // First of a stereo pair, skip next input pin
                     inIdx += 2;
                  }
                  else
                  {
                     // Mono input pin
                     inIdx += 1;
                  }

                  numIPorts++;

                  r.unsetFast();
               }
               else
               {
                  printf("[---] SyModuleVST::initVST: failed to query input pin #%d/%d\n", inIdx, numIn);

                  if((2 == numIn) && (0 == inIdx))
                  {
                     // Workaround for "Roland TB303 Screamer" plugin (does not support effGetInputProperties)
                     inIdx += 2;
                     numIPorts++;
                  }
                  else
                  {
                     YAC_DELETE_SAFE(plugin);

                     // Failed
                     return 0;
                  }
               }
            }
         }

         printf("xxx initVST 5\n");

         // Query output ports
         sSI numOut = plugin->getNumOutputs();
         sSI numOPorts = 0;

         sSI outIdx = 0;

         if(bPrealloc)
         {
            numOPorts = (sSI) num_outputs;
         }
         else
         {
            // Determine number of tksynth output ports
            while(outIdx < numOut)
            {
               // Query output pin
               YAC_Value r;
            
               plugin->getOutputProperties((sUI)outIdx, &r);

               if(r.type >= YAC_TYPE_OBJECT)
               {
                  VST2PinProperties *pin = (VST2PinProperties *) r.value.object_val;

                  if(pin->isStereo())
                  {
                     printf("xxx initVST: outIdx=%u isStereo\n", outIdx);
                     // First of a stereo pair, skip next output pin
                     outIdx += 2;
                  }
                  else
                  {
                     printf("xxx initVST: outIdx=%u isMono\n", outIdx);
                     // Mono output pin
                     outIdx += 1;
                  }

                  numOPorts++;

                  r.unsetFast();
               }
               else
               {
                  printf("[---] SyModuleVST::initVST: failed to query output pin #%d\n", outIdx);

                  YAC_DELETE_SAFE(plugin);

                  // Failed
                  return 0;
               }
            }
         }

         printf("xxx initVST 6\n");

         // Allocate and initialize the tksynth input ports
         if(!bPrealloc)
         {
            allocInputs(numIPorts);
         }

         printf("xxx initVST 7\n");

         inIdx = 0;
         sUI portInIdx = 0;
         while(inIdx < numIn)
         {
            // Query input pin
            YAC_Value r;
            
            plugin->getInputProperties((sUI)inIdx, &r); // (note) error checking has already been done above

            VST2PinProperties *pin = (VST2PinProperties *) r.value.object_val;

            sBool bStereo;
            const char *strLabel;

            YAC_Value rLabel;
            char buf[12];

            if(NULL != pin)
            {
               // Query input name (short label, label, or auto-generated string when the first two fail)
               pin->getShortLabel(&rLabel);
               if(IsEmptyString(&rLabel))
               {
                  pin->getLabel(&rLabel);
                  if(IsEmptyString(&rLabel))
                  {
                     sprintf(buf, "#%d", portInIdx);
                     strLabel = buf;
                  }
                  else
                  {
                     strLabel = (const char*)rLabel.value.string_val->chars;
                  }
               }
               else
               {
                  strLabel = (const char*)rLabel.value.string_val->chars;
               }

               printf("[dbg] SyModuleVST::initVST: input port [%u].name=\"%s\".\n", portInIdx, strLabel);

               bStereo = pin->isStereo();
            }
            else
            {
               if((2 == numIn) && (0==inIdx))
               {
                  // Workaround for plugins that do not support effGetInputProperties (e.g. "Roland TB303 Screamer")
                  bStereo = 1;
                  strLabel = "smp";
               }
            }

            SyInput *input;

            if(bPrealloc)
            {
               // Initialize input port
               input = initInput((sUI)portInIdx,
                                 strLabel,          // port name
                                 SYPORT_MODE_SE,    // sample-exact mode
                                 -1.0f, 1.0f, 0.0f  // min/max/def
                                 );
            }
            else
            {
               // Add input port
               input = addInput(strLabel,          // port name
                                SYPORT_MODE_SE,    // sample-exact mode
                                -1.0f, 1.0f, 0.0f  // min/max/def
                                );
            }

            input->num_channels = bStereo ? 2 : 1;

            printf("[dbg] SyModuleVST::initVST:  input[%u] #ch=%d name=\"%s\".\n", portInIdx, input->num_channels, (const char*)input->name->chars);

            rLabel.unsetFast();

            if(bStereo)
            {
               // First of a stereo pair, skip next input pin
               inIdx += 2;
            }
            else
            {
               // Mono input pin
               inIdx += 1;
            }

            r.unsetFast();

            // Next input port
            portInIdx++;
         }

         printf("xxx initVST 8 numOPorts=%u\n", numOPorts);

         // Allocate and initialize the tksynth output ports
         if(!bPrealloc)
         {
            allocOutputs(numOPorts);
         }

         printf("xxx initVST 9\n");

         outIdx = 0;
         sUI portOutIdx = 0;
         while(outIdx < numOut)
         {
            // Query output pin
            YAC_Value r;

            printf("xxx initVST 9a\n");
            
            plugin->getOutputProperties((sUI)outIdx, &r); // (note) error checking has already been done above

            printf("xxx initVST 9b\n");

            VST2PinProperties *pin = (VST2PinProperties *) r.value.object_val;

            // Query output name (short label, label, or auto-generated string when the first two fail)
            YAC_Value rLabel;
            const char *strLabel;
            char buf[12];
            pin->getShortLabel(&rLabel);
            if(IsEmptyString(&rLabel))
            {
               pin->getLabel(&rLabel);
               if(IsEmptyString(&rLabel))
               {
                  sprintf(buf, "#%d", portOutIdx);
                  strLabel = buf;
               }
               else
               {
                  strLabel = (const char*)rLabel.value.string_val->chars;
               }
            }
            else
            {
               strLabel = (const char*)rLabel.value.string_val->chars;
            }

            printf("[dbg] SyModuleVST::initVST: output port [%u].name=\"%s\".\n", portOutIdx, strLabel);

            printf("xxx initVST 9c\n");

            sBool bStereo = pin->isStereo();

            printf("xxx initVST 9d bStereo=%d bPrealloc=%d\n", bStereo, bPrealloc);

            SyOutput *output;
            if(bPrealloc)
            {
               // Initialize output port
               output = initOutput((sUI)portOutIdx,
                                   strLabel,         // port name
                                   SYPORT_MODE_SE    // sample-exact mode
                                   );
            }
            else
            {
               // Add output port
               output = addOutput(strLabel,         // port name
                                  SYPORT_MODE_SE    // sample-exact mode
                                  );
            }
            output->num_channels = bStereo ? 2 : 1;

            printf("xxx initVST 9e\n");

            printf("[dbg] SyModuleVST::initVST: output[%u] #ch=%d name=\"%s\".\n", portOutIdx, output->num_channels, (const char*)output->name->chars);

            rLabel.unsetFast();

            if(bStereo)
            {
               // First of a stereo pair, skip next output pin
               outIdx += 2;
            }
            else
            {
               // Mono output pin
               outIdx += 1;
            }

            r.unsetFast();

            // Next output port
            portOutIdx++;
         }

         printf("xxx initVST 10\n");

         // Activate
         plugin->resume();

         printf("xxx initVST 11\n");

         // Start processing
         plugin->startProcess();

         printf("xxx initVST 12\n");

         // Succeeded
         return 1;
      }
      else
      {
         printf("[---] SyModuleVST::initVST: failed to load plugin.\n");
      }
   }
   else
   {
      printf("[---] SyModuleVST::initVST: _pathName argument is not a String.\n");
   }
   
   // Failed
   return 0;
}

void SyModuleVST::allocPlugin(void) {
   if(NULL == plugin)
   {
      plugin = YAC_NEW(VST2Plugin);
   }
}

void SyModuleVST::reset(void) {
   SyModule::reset();
}

void SyModuleVST::doCalc(sUI _framesAdvance) {

   if(NULL != plugin)
   {

      ////printf("xxx SyModuleVST::doCalc: _framesAdvance=%u\n", _framesAdvance);

      // Process queued events
      plugin->processEvents();

      if(NULL != plugin->input_buffers)
      {
         // Copy current_value to mono/stereo sample VST buffer
         sUI inIdx;
         sUI vstInIdx = 0;
         for(inIdx=0; inIdx<num_inputs; inIdx++)
         {
            SyInput *in = inputs[inIdx];

            sUI i;
            for(i=0; i<in->num_channels; i++)
            {
               in->current_value[i] = plugin->input_buffers[vstInIdx][i];
            }

            vstInIdx += in->num_channels;
         }
      }

      // Process VST effect/instrument
      //  --> (ultra-slow) per-sample VST rendering mode. should never be used.
      doCalcBlock(1);

      if(NULL != plugin->output_buffers)
      {
         // Copy mono/stereo sample from port buffer to current_value
         sUI outIdx;
         sUI vstOutIdx = 0;
         for(outIdx=0; outIdx<num_outputs; outIdx++)
         {
            SyOutput *out = outputs[outIdx];

            sUI i;
            for(i=0; i<out->num_channels; i++)
            {
               out->current_value[i] = plugin->output_buffers[vstOutIdx+i][0] * out->level + out->offset;
               //printf("xxx out ch[%u]=%f\n", i, out->current_value[i]);
            }

            vstOutIdx += out->num_channels;
         }
      }

      ////printf("xxx SyModuleVST::doCalc: 5\n");
   }
}

void SyModuleVST::doCalcBlock(sUI _numFrames) {

   if(NULL != plugin)
   {
      // Process queued events
      plugin->processEvents();

      // Copy input buffers to plugin buffers, interleave L/R if necessary
      if(NULL != plugin->input_buffers)
      {
         sUI inIdx;
         sUI vstInIdx = 0;
         for(inIdx=0; inIdx<num_inputs; inIdx++)
         {
            SyInput *in = inputs[inIdx];

            sF32 *d = in->buffer;

            if(2 == in->num_channels)
            {
               // Stereo
            
               sF32 *vstInL = plugin->input_buffers[vstInIdx + 0];
               sF32 *vstInR = plugin->input_buffers[vstInIdx + 1];

               sUI i;
               for(i=0; i<_numFrames; i++)
               {
                  // Interleave L/R samples
                  vstInL[i] = *d++;
                  vstInR[i] = *d++;
                  //printf("xxx vstIn[%u]= %f %f\n", i, vstInL[i], vstInR[i]);
               }

               vstInIdx += 2;
            }
            else
            {
               // Mono
               sF32 *vstIn = plugin->input_buffers[vstInIdx];

               memcpy(vstIn, d, sizeof(sF32)*_numFrames);

               vstInIdx += 1;
            }

            // Next input port

         } // loop inputports
      }

      // Process VST effect/instrument
      plugin->processReplacing(_numFrames, 0u/*off*/);

      ////printf("xxx this=0x%p numIn=%u numOut=%u\n", this, num_inputs, num_outputs);

      //printf("%f %f\n", plugin->output_buffers[0][0], plugin->output_buffers[1][0]);
      
      // Copy VST output buffer(s) to port buffer(s)
      if(NULL != plugin->output_buffers)
      {
         sUI outIdx;
         sUI vstOutIdx = 0;
         for(outIdx=0; outIdx<num_outputs; outIdx++)
         {
            SyOutput *out = outputs[outIdx];

            sBool bNoScaling = (Dfltequal(out->offset, 0.0f) && Dfltequal(out->level, 1.0f));

            sF32 *d = out->buffer;

            if(2 == out->num_channels)
            {
               // Stereo

               sF32 *vstOutL = plugin->output_buffers[vstOutIdx + 0];
               sF32 *vstOutR = plugin->output_buffers[vstOutIdx + 1];

               if(bNoScaling)
               {
                  // Most frequent case: Do not offset/scale stereo samples
                  sUI i;
                  for(i=0; i<_numFrames; i++)
                  {
                     // Interleave stereo samples
                     ////printf("xxx vstOutLR[%u] = (%f, %f)\n", i, vstOutL[i], vstOutR[i]);
                     *d++ = vstOutL[i];
                     *d++ = vstOutR[i];
                  }
               }
               else
               {
                  // Apply offset/scale stereo samples
                  sUI i;
                  for(i=0; i<_numFrames; i++)
                  {
                     // Scale and interleave stereo samples
                     *d++ = vstOutL[i] * out->level + out->offset;
                     *d++ = vstOutR[i] * out->level + out->offset;
                  }
               }

               vstOutIdx += 2;
            }
            else
            {
               // Mono

               sF32 *vstOut = plugin->output_buffers[vstOutIdx];

               if(bNoScaling)
               {
                  // Most frequent case: Do not offset/scale mono samples

                  // (todo) change SyPort so it becames possible to copy the samples by reference (buffer_ptr member)

                  memcpy(d, vstOut, sizeof(sF32) * _numFrames);
               }
               else
               {
                  // Apply offset/scale mono samples

                  sUI i;
                  for(i=0; i<_numFrames; i++)
                  {
                     *d++ = vstOut[i] * out->level + out->offset;
                  }
               }

               vstOutIdx++;
            }

            // Next output

         } // loop outputs
      }

   } // if plugin

}

YAC_Object *SyModuleVST::getPlugin(void) {
   return plugin;
}

void SyModuleVST::uiSetVSTGeneratorId(sUI _id) {
   ui_vstgenerator_id = _id;
}

sUI SyModuleVST::uiGetVSTGeneratorId(void) {
   return ui_vstgenerator_id;
}

sBool SyModuleVST::queryPatchData(void) {

   if(NULL == patch_data)
   {
      patch_data = (YAC_Buffer*) YAC_New_Buffer();
   }

   sBool bDone = 0;
   
   if(plugin->getChunk(patch_data, 0 /* bank */))
   {
      patch_data_type = SYMODULEVST_DATATYPE_BANKCHUNK;

      printf("[dbg] SyModuleVST::queryPatchData: type=BANKCHUNK, size=%u\n", patch_data->size);
      
      bDone = 1;
   }
   else if(plugin->getChunk(patch_data, 1 /* program */))
   {
      patch_data_type = SYMODULEVST_DATATYPE_PROGRAMCHUNK;

      printf("[dbg] SyModuleVST::queryPatchData: type=PROGRAMCHUNK, size=%u\n", patch_data->size);
      
      bDone = 1;
   }
   
   if(!bDone)
   {
      // Plugin does not support chunks, save parameter array
      
      patch_data_type = SYMODULEVST_DATATYPE_PARAMETERS;
      
      sUI numParams = (sUI) plugin->getNumParameters();
      
      patch_data->yacArrayRealloc(numParams * sizeof(sF32), 0,0,0);
      
      patch_data->io_offset = 0;
      
      for(sUI i=0; i<numParams; i++)
      {
         patch_data->yacStreamWriteF32(plugin->getParameter((sSI)i));
      }

      printf("[dbg] SyModuleVST::queryPatchData: type=PARAMETERS, size=%u\n", patch_data->size);
   }

   return 1;
}

sBool SyModuleVST::restorePatchData(void) {
   if(NULL != plugin)
   {
      if(NULL != patch_data)
      {
         switch(patch_data_type)
         {
            default:
            {
               printf("[---] SyModuleVST::readStream: unknown patch_data_type(%d). unable to restore patch data.\n", patch_data_type);
            }
            break;
            
            case SYMODULEVST_DATATYPE_BANKCHUNK:
            {
               printf("[dbg] SyModuleVST::restorePatchData: type=BANKCHUNK, size=%u\n", patch_data->size);

               plugin->setChunk(patch_data, 0 /*bank*/);
            }
            return 1;
               
            case SYMODULEVST_DATATYPE_PROGRAMCHUNK:
            {
               printf("[dbg] SyModuleVST::restorePatchData: type=PROGRAMCHUNK, size=%u\n", patch_data->size);

               plugin->setChunk(patch_data, 1 /*program*/);
            }
            return 1;
               
            case SYMODULEVST_DATATYPE_PARAMETERS:
            {
               printf("[dbg] SyModuleVST::restorePatchData: type=PARAMETERS, size=%u\n", patch_data->size);

               sUI numParams = plugin->getNumParameters();
               sUI numRestoreParams = patch_data->size / sizeof(sF32);
               if(numParams != numRestoreParams)
               {
                  printf("[~~~] SyModuleVST::restorePatchData: numParams != numRestoreParams (%u != %u)\n", numParams, numRestoreParams);
               }
               sUI i;
               sF32 *s = (sF32*) patch_data->buffer;
               for(i=0; i<numRestoreParams; i++)
               {
                  plugin->setParameter(i, *s++);
               }
            }
            return 1;
         }
      }
      else
      {
         printf("[---] SyModuleVST::restorePatchData: patch_data is NULL.\n");
      }
   }
   else
   {
      printf("[---] SyModuleVST::restorePatchData: plugin is NULL.\n");
   }

   // Failed
   return 0;
}


sBool SyModuleVST::saveStream(YAC_Object *_ofs) {
   // Write version
   _ofs->yacStreamWriteI16(0x0002);

   // Write ui_vstgenerator_id
   _ofs->yacStreamWriteI16((sS16) ui_vstgenerator_id);

   // Write #input ports (placeholder until VST has actually been loaded)
   _ofs->yacStreamWriteI16((sS16) num_inputs);

   // Write basic input port information (placeholder until VST has actually been loaded)
   {
      sUI i;
      for(i=0; i<num_inputs; i++)
      {
         SyInput *inp = inputs[i];

         // Input name
         _ofs->yacStreamWriteString(inp->name, 0, 256);

         // Next input port
      }
   }

   // Write #output ports (placeholder until VST has actually been loaded)
   _ofs->yacStreamWriteI16((sS16) num_outputs);

   // Write basic output port information (placeholder until VST has actually been loaded)
   {
      sUI i;
      for(i=0; i<num_outputs; i++)
      {
         SyOutput *outp = outputs[i];

         // Output name
         _ofs->yacStreamWriteString(outp->name, 0, 256);

         // Next output port
      }
   }

   // Write patch data
   {
      queryPatchData();

      _ofs->yacStreamWriteI8((sS8) patch_data_type);

      _ofs->yacStreamWriteI32((sSI)patch_data->size);

      _ofs->yacStreamWriteBuffer(patch_data, 0, patch_data->size);
   }

   return 1;
}

sBool SyModuleVST::readStream(YAC_Object *_ifs) {
   // Read version
   sU16 v =_ifs->yacStreamReadI16();
   if(v >= 0x0001)
   {
      // Read ui_vstgenerator_id
      ui_vstgenerator_id = (sUI) (_ifs->yacStreamReadI16() & 65535);


      // Read #input ports (placeholder until VST has actually been loaded)
      num_inputs = (sUI) _ifs->yacStreamReadI16();

      // Read basic input port information (placeholder until VST has actually been loaded)

      allocInputs(num_inputs);
      {
         sUI i;
         for(i=0; i<num_inputs; i++)
         {
            SyInput *inp = inputs[i];
            // Read input name
            _ifs->yacStreamReadString(inp->name, 256);

            // Next input port
         }
      }

      // Read #output ports (placeholder until VST has actually been loaded)
      num_outputs = (sUI) _ifs->yacStreamReadI16();

      // Read basic output port information (placeholder until VST has actually been loaded)
      allocOutputs(num_outputs);
      {
         sUI i;
         for(i=0; i<num_outputs; i++)
         {
            SyOutput *outp = outputs[i];
            // Read output name
            _ifs->yacStreamReadString(outp->name, 256);
            ////printf("xxx SyModuleVST::readStream: output[%u].name=\"%s\".\n", i, outp->name->chars);

            // Next output port
         }
      }

      if(v >= 2)
      {
         // Read patch data
         if(NULL == patch_data)
         {
            patch_data = (YAC_Buffer*) YAC_New_Buffer();
         }

         patch_data_type = _ifs->yacStreamReadI8();

         sUI patchDataSize = (sUI) _ifs->yacStreamReadI32();
         
         _ifs->yacStreamReadBuffer(patch_data, 0, patchDataSize, 1/*resize*/);

         // (note) cannot restore patch data at this point because the plugin has not been loaded, yet.
      }

      // Succeeded
      return 1;

   } // if version check

   // Failed
   return 0;
}

